@@ -222,7 +222,7 @@ soup_buffer_get_owner (SoupBuffer *buffer)
 {
 	SoupBufferPrivate *priv = (SoupBufferPrivate *)buffer;
 
-	g_return_val_if_fail (priv->use == SOUP_MEMORY_OWNED, NULL);
+	g_return_val_if_fail ((int)priv->use == (int)SOUP_MEMORY_OWNED, NULL);
 	return priv->owner;
 }
 
@@ -527,6 +527,7 @@ soup_message_body_truncate (SoupMessageBody *body)
 		soup_buffer_free (iter->data);
 	g_slist_free (priv->chunks);
 	priv->chunks = priv->last = NULL;
+	priv->base_offset = 0;
 
 	if (priv->flattened) {
 		soup_buffer_free (priv->flattened);
