@@ -66,6 +66,8 @@ typedef struct {
 	guint        num_conns;
 
 	guint        num_messages;
+
+	SoupHTTPVersion http_version;
 } SoupSessionHost;
 
 typedef struct {
@@ -109,6 +111,7 @@ static void cancel_message  (SoupSession *session, SoupMessage *msg,
 			     guint status_code);
 static void auth_required   (SoupSession *session, SoupMessage *msg,
 			     SoupAuth *auth, gboolean retrying);
+static void flush_queue     (SoupSession *session);
 
 static void auth_manager_authenticate (SoupAuthManager *manager,
 				       SoupMessage *msg, SoupAuth *auth,
@@ -255,6 +258,7 @@ soup_session_class_init (SoupSessionClass *session_class)
 	session_class->requeue_message = requeue_message;
 	session_class->cancel_message = cancel_message;
 	session_class->auth_required = auth_required;
+	session_class->flush_queue = flush_queue;
 
 	/* virtual method override */
 	object_class->dispose = dispose;
@@ -1033,6 +1037,7 @@ soup_session_host_new (SoupSession *session, SoupURI *uri)
 	host = g_slice_new0 (SoupSessionHost);
 	host->uri = soup_uri_copy_host (uri);
 	host->addr = soup_address_new (host->uri->host, host->uri->port);
+	host->http_version = SOUP_HTTP_1_1;
 
 	return host;
 }
@@ -1185,6 +1190,8 @@ soup_session_send_queue_item (SoupSession *session,
 			      SoupMessageCompletionFn completion_cb)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupHTTPVersion http_version;
+	SoupSessionHost *host;
 
 	if (priv->user_agent) {
 		soup_message_headers_replace (item->msg->request_headers,
@@ -1199,6 +1206,28 @@ soup_session_send_queue_item (SoupSession *session,
 					     priv->accept_language);
 	}
 
+	g_mutex_lock (priv->host_lock);
+	host = get_host_for_message (session, item->msg);
+	http_version = host->http_version;
+	g_mutex_unlock (priv->host_lock);
+
+	/* Force keep alive connections for HTTP 1.0. Performance will
+	 * improve when issuing multiple requests to the same host in
+	 * a short period of time, as we wouldn't need to establish
+	 * new connections. Keep alive is implicit for HTTP 1.1.
+	 */
+	if (http_version == SOUP_HTTP_1_0) {
+		const gchar* conn_header;
+
+		conn_header = soup_message_headers_get_list (item->msg->request_headers, "Connection");
+		if (!conn_header ||
+		    (!soup_header_contains (conn_header, "Keep-Alive") &&
+		     !soup_header_contains (conn_header, "close")))
+
+			soup_message_headers_append (item->msg->request_headers,
+						     "Connection", "Keep-Alive");
+	}
+
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
 		       item->msg, soup_connection_get_socket (item->conn));
 	soup_connection_send_request (item->conn, item, completion_cb, item);
@@ -1288,7 +1317,6 @@ soup_session_make_connect_message (SoupSession    *session,
 	item->conn = g_object_ref (conn);
 	g_object_unref (msg);
 
-	item->conn = g_object_ref (conn);
 	g_signal_emit (session, signals[TUNNELING], 0, conn);
 	return item;
 }
@@ -1461,13 +1489,37 @@ soup_session_set_item_status (SoupSession          *session,
 			g_free (msg);
 			break;
 		}
-		/* else fall through */
+		soup_message_set_status (item->msg, status_code);
+		break;
+
+	case SOUP_STATUS_SSL_FAILED:
+		if (!g_tls_backend_supports_tls (g_tls_backend_get_default ())) {
+			soup_message_set_status_full (item->msg, status_code,
+						      "TLS/SSL support not available; install glib-networking");
+		} else
+			soup_message_set_status (item->msg, status_code);
+		break;
 
 	default:
 		soup_message_set_status (item->msg, status_code);
+		break;
 	}
 }
 
+static void
+got_headers_cb (SoupMessage *msg, gpointer data)
+{
+	SoupMessageQueueItem *item = data;
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (item->session);
+	SoupSessionHost *host;
+
+	/* Update the HTTP version to use for messages to this host */
+	g_mutex_lock (priv->host_lock);
+	host = get_host_for_message (item->session, SOUP_MESSAGE (msg));
+	host->http_version = soup_message_get_http_version (msg);
+	g_mutex_unlock (priv->host_lock);
+}
+
 static void
 queue_message (SoupSession *session, SoupMessage *msg,
 	       SoupSessionCallback callback, gpointer user_data)
@@ -1489,6 +1541,9 @@ queue_message (SoupSession *session, SoupMessage *msg,
 			G_CALLBACK (redirect_handler), item);
 	}
 
+	/* Used to keep track of the HTTP version used by the host */
+	g_signal_connect (msg, "got-headers", G_CALLBACK (got_headers_cb), item);
+
 	g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
 }
 
@@ -1633,14 +1688,8 @@ cancel_message (SoupSession *session, SoupMessage *msg, guint status_code)
 	item = soup_message_queue_lookup (priv->queue, msg);
 	g_return_if_fail (item != NULL);
 
-	if (item->cancellable)
-		g_cancellable_cancel (item->cancellable);
-
 	soup_message_set_status (msg, status_code);
-	if (soup_message_io_in_progress (msg))
-		soup_message_io_finished (msg);
-	else
-		item->state = SOUP_MESSAGE_FINISHING;
+	g_cancellable_cancel (item->cancellable);
 
 	soup_message_queue_item_unref (item);
 }
@@ -1704,6 +1753,20 @@ gather_conns (gpointer key, gpointer host, gpointer data)
 	*conns = g_slist_prepend (*conns, g_object_ref (conn));
 }
 
+static void
+flush_queue (SoupSession *session)
+{
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupMessageQueueItem *item;
+
+	for (item = soup_message_queue_first (priv->queue);
+	     item;
+	     item = soup_message_queue_next (priv->queue, item)) {
+		soup_session_cancel_message (session, item->msg,
+					     SOUP_STATUS_CANCELLED);
+	}
+}
+
 /**
  * soup_session_abort:
  * @session: the session
@@ -1714,18 +1777,12 @@ void
 soup_session_abort (SoupSession *session)
 {
 	SoupSessionPrivate *priv;
-	SoupMessageQueueItem *item;
 	GSList *conns, *c;
 
 	g_return_if_fail (SOUP_IS_SESSION (session));
 	priv = SOUP_SESSION_GET_PRIVATE (session);
 
-	for (item = soup_message_queue_first (priv->queue);
-	     item;
-	     item = soup_message_queue_next (priv->queue, item)) {
-		soup_session_cancel_message (session, item->msg,
-					     SOUP_STATUS_CANCELLED);
-	}
+	SOUP_SESSION_GET_CLASS (session)->flush_queue (session);
 
 	/* Close all connections */
 	g_mutex_lock (priv->host_lock);
